/*
Copyright (c) 2011, Julian Lyndon-Smith (julian+maia@dotr.com)
http://www.dotr.com
All rights reserved.

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction except as noted below, including without limitation
the rights to use,copy, modify, merge, publish, distribute,
and/or sublicense, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

 The Software and/or source code cannot be copied in whole and
 sold without meaningful modification for a profit.

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in
 the documentation and/or other materials provided with
 the distribution.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

using Progress.Lang.*.
using dotr.Maia.Model.*.

/** Maia crud builder 
 * extends the basic builder to add methods to build crud files from the crud.i template provided
 * @author Julian Lyndon-Smith (julian+maia@dotr.com) 
 */

class dotr.Maia.Builder.BuildActiveRecordBase inherits dotr.Maia.Builder.Basic: 
  constructor BuildActiveRecordBase (p_ObjectConfig as ObjectConfig):
    def var ObjectBuilder as dotr.Maia.Builder.ObjectBuilder no-undo.
    
    super(p_ObjectConfig).
    
    assign ObjectBuilder = dotr.Maia.Builder.ObjectBuilder:Instance.
    
    assign BuildPackage   =  substitute("&1.Model.Base",dotr.Maia.CurrentProject:Name)
           BuildDirectory = substitute("&1/Model/Base",dotr.Maia.CurrentProject:Path).
  end constructor.

  /** generate next-value statements
   * loop through all properties configured with a sequence and generate a next-value statement
   * only set the property if it is currently unknown or 0
   * @param p_Directive : holds the directive name and all parameters
   */

  method public void NextValue(p_Directive as Directive):
    def var lv_Data as char no-undo.
    
    def var Property1 as Property no-undo.
    
    for each TTProperty no-lock:
      Property1 = cast(TTProperty.property,Property).
       
      /** is there a sequence defined for this property */
      if Property1:Sequence eq "" or Property1:Sequence eq "-- Not Used --" then next.
       
      assign lv_Data = lv_Data + substitute("if this-object:&1 eq 0 or this-object:&1 eq ? then assign this-object:&1  = next-value(&2,<<DB>>).",
                                             TTProperty.PropertyName,
                                             Property1:Sequence).
    end.
    assign p_Directive:Data = trim(lv_Data,"~n").                                                                                                                                                  
  end method.
  
  /** generate guid statements
   * loop through all properties configured with a generateGUID flag and generate a next-value statement
   * only set the property if it is currently unknown or ""
   * @param p_Directive : holds the directive name and all parameters
   */

  method public  void GenerateGUID(p_Directive as Directive):
    def var lv_Data as char no-undo.
    
    for each TTProperty no-lock:
      if cast(TTProperty.property,Property):GenerateGUID then 
          assign lv_Data = lv_Data + substitute("~nif this-object:&1 eq '' or this-object:&1 eq ? then assign this-object:&1  = guid(generate-uuid).",
                                                TTProperty.PropertyName).
    end.
    assign p_Directive:Data = trim(lv_Data,"~n").
  end method.
  
  /** generate assign to buffer statement
   *
   *  loop through all properties for an object, and build the appropriate assign to buffer statement
   *  @param p_Directive : holds the directive name and all parameters
   */

  method public void ActiveRecordModel_AssignToBuffer(p_Directive as Directive):
    def var lv_Data   as char no-undo.
    def var lv_Assign as char no-undo.
    
    def var lv_maxLen as int no-undo.
    def var lv_i      as int no-undo.
    
    def var Property1 as Property no-undo.
    
    /** get max size of property name. Used to align code properly */
    for each TTProperty no-lock :
      Property1 = cast(TTProperty.Property,Property).

      /** a manual property has no db relation , so skip it*/
      if Property1:ManualProperty then next.
      
      if Property1:BuildProperty then assign lv_MaxLen = MAX(lv_MaxLen,if Property1:extent gt 0 then length(Property1:FieldName) + 2 + length(string(Property1:extent))
                                                                                                else length(Property1:FieldName)).
    end.
    
    /** now let's loop through, building the assign statement */
    for each TTProperty no-lock:
      Property1 = cast(TTProperty.Property,Property).
      
      if not Property1:BuildProperty then next.

      /** a manual property has no db relation , so skip it*/
      if Property1:ManualProperty then next.
      
      if Property1:extent gt 0 then
      do lv_i = 1 to Property1:extent:
        assign lv_Assign = substitute("<<TableName>>.&2[&3]@ = this-object:&1[&3]~n",Property1:PropertyName,
                                                                                          Property1:FieldName,
                                                                                          lv_i)
               lv_Data   = lv_Data + replace(lv_Assign,"@",fill(" ",lv_maxLen - length(Property1:FieldName))).
      end.
      
      else /** not an extent */
      do:
        assign lv_Assign = substitute("<<TableName>>.&2@ = this-object:&1~n",Property1:PropertyName,
                                                                                  Property1:FieldName)
               lv_Data   = lv_Data + replace(lv_Assign,"@",fill(" ",lv_maxLen - length(Property1:FieldName))).
      end.
                      
    end.
    
    assign p_Directive:Data = trim(lv_Data,"~n") + ".".                                                                                                                                                  
  end method.

  /** create statement to assign database fields to properties
   * @param p_Directive : holds the directive name and all parameters
   */
  
  method public void ActiveRecordModel_AssignFromBuffer(p_Directive as Directive):
    def var lv_Data   as char no-undo.
    def var lv_Assign as char no-undo.
    
    def var lv_maxLen as int no-undo.
    
    def var Property1 as Property no-undo.
    
    /** get max size of property name. Used to align code properly */
    for each TTProperty no-lock :
      Property1 = cast(TTProperty.Property,Property).

      /** a manual property has no db relation , so skip it*/
      if Property1:ManualProperty then next.
      
      if Property1:BuildProperty then assign lv_MaxLen = MAX(lv_MaxLen,TTProperty.SIZE).
    end.
    
    /** now let's loop through, building the assign statement */
    for each TTProperty no-lock:
      Property1 = cast(TTProperty.Property,Property).
      
      if not Property1:BuildProperty then next.

      /** a manual property has no db relation , so skip it*/
      if Property1:ManualProperty then next.
      
      assign lv_Assign = substitute("this-object:&1@ = buffer::&2~n",Property1:PropertyName,
                                                                              Property1:FieldName)
             lv_Data   = lv_Data + replace(lv_Assign,"@",fill(" ",lv_maxLen - TTProperty.Size)). 
    end.
    
    assign lv_Assign = "this-object:rowid@ = buffer:rowid~n"
           lv_Data   = lv_Data + replace(lv_Assign,"@",fill(" ",lv_maxLen - length("rowid"))). 

    assign p_Directive:Data = trim(lv_Data,"~n") + ".".                                                                                                                                                  
  end method.

 /** insert fk methods
  *
  *  @param p_template  : holds the template to use
  *  @param p_Directive : holds the directive name and all parameters
  */

  method private longchar ActiveRecord_Insertfk(p_template as char, p_Directive as Directive):
    def var lv_Data    as char        no-undo.
    
    def var lv_i as int no-undo.

    def var lv_Header as longchar no-undo.
    
    /** check for custom builder file */
    assign file-info:file-name = if GetParameter(p_Directive,"Source") eq "" then substitute("dotr/Maia/Template/builder/&1",p_template)
                                                                             else GetParameter(p_Directive,"Source").                                  
    
    if file-info:full-pathname eq ? then assign lv_header = substitute("/** template &1 not found */",p_template). 
    else copy-lob from file file-info:full-pathname to lv_Header.
    
    do lv_i = 1 to ObjectConfig1:NumProperties:
             
      if ObjectConfig1:Property[lv_i]:isForeignKey then
      do:
        
        assign lv_Data = lv_Data + replace(lv_Header,"<<Cascademethod>>",if ObjectConfig1:Property[lv_i]:fkUpdate 
                                                                            then "Assign<<fkObject>>('')"
                                                                            else "remove()")
               lv_Data = replace(lv_Data,"<<fkFieldName>>",ObjectConfig1:Property[lv_i]:FieldName)
               lv_Data = replace(lv_Data,"<<fkProperty>>",ObjectConfig1:Property[lv_i]:PropertyName)
               lv_Data = replace(lv_Data,"<<fkObject>>",ObjectConfig1:Property[lv_i]:ForeignKey:PropertyName).
      end.             
    end.      

    return lv_data.
  end method.

 /** updates fk on deletion of fk 
  *
  *  loops through all objects with a fk of this object, and sets the fk property to "" 
  *  @param p_Directive : holds the directive name and all parameters
  */

  method public void ActiveRecordHelper_fkDeleted(p_Directive as Directive):  /** when a fk object is deleted, need to either remove or update related object (update orderline when order deleted) */
    assign p_Directive:Data = ActiveRecord_Insertfk("ActiveRecordHelper_fkDeleted.i",p_Directive).
  end method.  

   /** assign a fk to this model
   *
   *  for each fk of this model, create a method to assign the fk to the model
   *
   *  @param p_Directive : holds the directive name and all parameters
   */

  method public void ActiveRecordModel_fkAssign(p_Directive as Directive):
    assign p_Directive:Data = ActiveRecord_Insertfk("ActiveRecordModel_fkAssign.i",p_Directive).
  end method.  
  
  /** Create Cascade statements
  *
  *  when a record is deleted, we need to either remove all associated records (remove) or 
  *  remove the fk reference. This is done by calling the appropriate helper methods in the target object class
  *  @param p_Directive : holds the directive name and all parameters
  */

  method public void ActiveRecordDelete(p_Directive as Directive): /** when deleteing record, update all associated records with the fk (Delete order, remove orderlines */
    def var lv_Cascade as char extent no-undo.
    def var lv_Data    as char        no-undo.
    def var lv_List    as char        no-undo init "update,delete".
    
    def var fk1 as dotr.Maia.Model.fkModel no-undo.
    def var CascadeObjectConfig as dotr.Maia.Model.ObjectConfig no-undo.
    
    def var lv_i as int no-undo.
    def var lv_j as int no-undo.
    def var lv_k as int no-undo.
    
    fk1 = new dotr.Maia.Model.fkModel().
    
    do lv_j = 1 to num-entries(lv_List):
      extent(lv_cascade) = ?.
      
      assign lv_Cascade = fk1:Getfk(ObjectConfig1:ClassName,entry(lv_j,lv_List)).
      
      do lv_i = 1 to extent(lv_Cascade):
        
        if lv_Cascade[lv_i] eq "" then next.
        
        CascadeObjectConfig = dotr.Maia.Model.ObjectConfig:Get(lv_Cascade[lv_i]).
        if not valid-object(CascadeObjectConfig) then next.

        do lv_k = 1 to CascadeObjectConfig:NumProperties:
          if CascadeObjectConfig:Property[lv_k]:isForeignKey and CascadeObjectConfig:Property[lv_k]:fkClassName eq ObjectConfig1:ClassName then
          do:
            assign lv_Data = lv_Data + substitute("&1.&2:&3Deleted(this-object:<<KeyProperty>>).~n",
                                                replace(CascadeObjectConfig:ModelPackage,"Model","Helper"),
                                                CascadeObjectConfig:ModelName,
                                                CascadeObjectConfig:Property[lv_k]:ForeignKey:PropertyName).
          end.             
        end.        
      end.
    end.  
    
    assign p_Directive:Data = trim(lv_Data,"~n").
     
  end method.
  
 /** opens a query on all records of this model that have this fk
  *
  *  for each fk of this model, create a helper method that loops through all records with this fk
  *
  *  @param p_Directive : holds the directive name and all parameters
  */

  method public void ActiveRecordHelper_fkGet(p_Directive as Directive):
    assign p_Directive:Data = ActiveRecord_Insertfk("ActiveRecordHelper_fkGet.i",p_Directive).
  end method.  

  /** build the methods for each of the get statements in the model
   * @param p_Directive : holds the directive name and all parameters
   */
  
  method public void ActiveRecordHelper_GetMethods(p_Directive as Directive):
    def var lv_i      as int no-undo.

    def var lv_Index as char no-undo.

    def var lv_Header as longchar no-undo.
    
    def var Index1 as dotr.Maia.Model.Index extent no-undo.
    
    /** check for custom builder file */
    assign file-info:file-name = if GetParameter(p_Directive,"Source") eq "" then "dotr/Maia/Template/builder/getModelMethod.i"
                                                                           else GetParameter(p_Directive,"Source").                                  
    
    copy-lob from file file-info:full-pathname to lv_Header.
    
    Index1 = (new dotr.Maia.Builder.DBProperties()):BuildIndex(ObjectConfig1).

    do lv_i = 1 to extent(Index1):
      if Index1[lv_i]:IndexName eq "" then next.
      
      assign lv_Index = replace(lv_Header,"<<GetName>>",Index1[lv_i]:IndexName)
             lv_index = replace(lv_index,"<<DocParameterList>>",replace(Index1[lv_i]:parametersIn," as ", " - "))
             lv_index = replace(lv_index,"<<FindString>>",Index1[lv_i]:FindString)
             lv_index = replace(lv_index,"<<ParameterList>>",Index1[lv_i]:parametersIn)
             lv_index = replace(lv_index,"<<ParameterCall>>",Index1[lv_i]:parametersCall)
    
             p_Directive:Data = p_Directive:Data + "~n" + lv_Index.
    end.
    
  end method.


end class.