/*
Copyright (c) 2011, Julian Lyndon-Smith (julian+maia@dotr.com)
http://www.dotr.com
All rights reserved.

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction except as noted below, including without limitation
the rights to use,copy, modify, merge, publish, distribute,
and/or sublicense, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

 The Software and/or source code cannot be copied in whole and
 sold without meaningful modification for a profit.

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in
 the documentation and/or other materials provided with
 the distribution.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

using Progress.Lang.*.
using dotr.Maia.Model.*.

/** Maia lookup builder 
 * extends the basic builder to add methods to build lookup user control files from the lookup.i template provided
 * @author Julian Lyndon-Smith (julian+maia@dotr.com) 
 */

class dotr.Maia.Builder.BuildLookup inherits dotr.Maia.Builder.Basic:
  def private var NumFields as int no-undo. /** number of fields defined to be shown in the lookup */
  
  def private var DisplayField as char no-undo. /** the field to be shown in the target text */
  
  constructor BuildLookup (p_ObjectConfig as ObjectConfig):
    def var ObjectBuilder as dotr.Maia.Builder.ObjectBuilder no-undo.
    
    super(p_ObjectConfig).
    
    assign ObjectBuilder = dotr.Maia.Builder.ObjectBuilder:Instance.
    
    assign BuildPackage   = substitute("&1.UI.winform.Lookup",dotr.Maia.CurrentProject:Name)
           BuildDirectory = substitute("&1/UI/winform/Lookup",dotr.Maia.CurrentProject:Path).
   
    Init().
  end constructor.

  /** find the number of fields to show, and get the key display field */
 
  method private void init():  

    for each TTProperty no-lock where TTProperty.DisplayInLookup gt 0 NumFields = 1 to NumFields + 1:
      if DisplayField eq "" then assign DisplayField = TTProperty.FieldName.               
    end.
    
    assign BuilderAvailable = numFields gt 0 . 
    
  end method.

  /** create the define statement for the columns needed for the lookup grid 
   * @param p_Directive : holds the directive name and all parameters
   */

  method public void DefineDisplayColumns(p_Directive as Directive):

    def var lv_Data as char no-undo.
    
    if NumFields eq 0 then /** nothing to build, return */ 
    do:
     assign p_Directive:Data = "".
     return.
    end.

    /** loop through all lookup fields */    
    for each TTProperty no-lock where TTProperty.DisplayInLookup gt 0 by TTProperty.DisplayInLookup :
      assign lv_Data = lv_Data + substitute("def private var col&1 as DevExpress.XtraGrid.Columns.GridColumn no-undo.~n",TTProperty.PropertyName).  
    end.
    
    assign p_Directive:Data = trim(lv_Data,"~n").
  end method.

  /** create the "new" statement for the columns needed for the lookup grid 
   * @param p_Directive : holds the directive name and all parameters
   */

  method public void NewGridColumns(p_Directive as Directive):

    def var lv_Data as char no-undo.

    if NumFields eq 0 then /** nothing to do */ 
    do:
     assign p_Directive:Data = "".
     return.
    end.
    
    for each TTProperty no-lock where TTProperty.DisplayInLookup gt 0 by TTProperty.DisplayInLookup :
      assign lv_Data = lv_Data + substitute("this-object:col&1 = new DevExpress.XtraGrid.Columns.GridColumn().~n",TTProperty.PropertyName).  
    end.
    
    assign p_Directive:Data = trim(lv_Data,"~n").
  end method.

  /** create the statements required to generate the data schema for the data source
   * @param p_Directive : holds the directive name and all parameters
   */

  method public void DataSchema(p_Directive as Directive):
    def var lv_Data as char no-undo.
    
    def var lv_i as int no-undo.

    if NumFields eq 0 then /** nothing to do */
    do:
     assign p_Directive:Data = "".
     return.
    end.
    
    /** always add the keyfield name in as a column so that we have a unique value to work with */
    assign lv_Data = substitute("def var arrayvar2 AS Progress.Data.ColumnPropDesc EXTENT &1 NO-undo.~n",NumFields + 1) +
                     'arrayvar2[1] = new Progress.Data.ColumnPropDesc("<<KeyFieldName>>", "<<KeyFieldName>>", Progress.Data.DataType:CHARACTER).~n'.
                     
    /** as we've already added one column (the keyfield), we need to increase the column count by one */  
    for each TTProperty no-lock where TTProperty.DisplayInLookup gt 0 by TTProperty.DisplayInLookup lv_i = 2 to lv_i + 1:
      assign lv_Data = lv_Data + substitute('arrayvar2[&1] = new Progress.Data.ColumnPropDesc("&2", "&3", Progress.Data.DataType:CHARACTER).~n',
                                            lv_i,
                                            TTProperty.PropertyName,
                                            TTProperty.FieldName).  
    end.
      
    assign lv_Data = lv_Data + "tableDesc1:Columns = arrayvar2.~n" +
                               substitute('this-object:gridLookUpEdit1:Properties:DisplayMember = "&1".~n',DisplayField) .
    
    assign p_Directive:Data = trim(lv_Data,"~n").
  end method.

  /** generate the statement required to add ranges to the grid
   * @param p_Directive : holds the directive name and all parameters
   */

  method public void AddGridColumns(p_Directive as Directive):
    def var lv_Data as char no-undo.
    
    def var lv_i as int no-undo.

    if NumFields eq 0 then /** nothing to do */ 
    do:
     assign p_Directive:Data = "".
     return.
    end.
    
    assign lv_Data = substitute("def var LookupArray as DevExpress.XtraGrid.Columns.GridColumn extent &1 no-undo.~n",NumFields).

    for each TTProperty no-lock where TTProperty.DisplayInLookup gt 0 by TTProperty.DisplayInLookup lv_i = 1 to lv_i + 1:
      assign lv_Data = lv_Data + substitute("LookupArray[&1] = this-object:col&2.~n",lv_i,TTProperty.PropertyName).  
    end.
      
    assign lv_Data = lv_Data + "this-object:gridLookUpEdit1View:Columns:AddRange(LookupArray).".
    
    assign p_Directive:Data = trim(lv_Data,"~n").
  end method.

  /** generate the statement required to show the columns in the grid
   * @param p_Directive : holds the directive name and all parameters
   */

  method public void ViewDisplayColumns(p_Directive as Directive):
    def var lv_Data as char no-undo.
    
    def var lv_i as int no-undo.

    if NumFields eq 0 then /** nothing to do */ 
    do:
     assign p_Directive:Data = "".
     return.
    end.
    
    for each TTProperty no-lock where TTProperty.DisplayInLookup gt 0 by TTProperty.DisplayInLookup lv_i = 0 to lv_i + 1:
      assign lv_Data = lv_Data + substitute('/*  */~n' +
                                            '/* col&1 */~n' +
                                            '/*  */~n' +
                                            'THIS-OBJECT:col&1:Caption = "&1".~n' +
                                            'THIS-OBJECT:col&1:FieldName = "&2".~n' +
                                            'THIS-OBJECT:col&1:Name = "col&1".~n' +
                                            'THIS-OBJECT:col&1:Visible = TRUE.~n' +
                                            'THIS-OBJECT:col&1:VisibleIndex = &3.~n',
                                            
                                            TTProperty.PropertyName,
                                            TTProperty.FieldName,
                                            TTProperty.DisplayInLookup - 1).
    end.
      
    assign p_Directive:Data = trim(lv_Data,"~n").
  end method.
  

end class.