/*
Copyright (c) 2011, Julian Lyndon-Smith (julian+maia@dotr.com)
http://www.dotr.com
All rights reserved.

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction except as noted below, including without limitation
the rights to use,copy, modify, merge, publish, distribute,
and/or sublicense, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

 The Software and/or source code cannot be copied in whole and
 sold without meaningful modification for a profit.

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in
 the documentation and/or other materials provided with
 the distribution.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

using Progress.Lang.*.
using dotr.Maia.Model.*.

/** crud library for objectConfig 
 * this class is responsible for all crud activity on the ObjectConfig model. It serialises and deserialises the 
 * config model and all related children (templates, properties, custom code etc). This data is stored in the 
 * Maia Config/<<ClassName>>.xml file
 *
 * @author Julian Lyndon-Smith (julian+maia@dotr.com) 
 */

routine-level on error undo, throw.

class dotr.Maia.Crud.ObjectConfig:

  /** temptable defs for the ObjectConfig class itself */
  def temp-table TTObjectConfig no-undo xml-node-name "ObjectConfig"     
    field ClassName   as char /* name of the class (base name only) [foo] */
    field TableName   as char /* table name of the class (base name only) [foo] */
    field KeyFieldName      as char 
    field KeyProperty as char 
    field KeyFieldDatatype as char 
    field modelName as char
    field modelPackage as char
    field fkDelete as char
    field LogicalDB as char
    field BuildObject  as logical init yes /* do we build this class or not */
    field isAbstract   as logical init no 
    field CRC           as int
    
    index main is primary unique 
      ClassName 
    .

  /** temptable defs for the templates used by the ObjectConfig class */
  def temp-table TTTemplate no-undo xml-node-name "Templates"
    field ClassName    as char
    field TemplateName as char
    field MD5Hash as char
    field buildTemplate as logical init yes 
    field buildIfSame   as logical
    field buildIfExists as logical
      
    index Main is primary unique
      ClassName
      TemplateName
      . 

  /** temptable defs for the custom code used by the ObjectConfig class */
  def temp-table TTCustomCode no-undo xml-node-name "CustomCode"
    field ClassName    as char
    field CodeName    as char
    field Data        as CLOB
    
    index Main is primary unique
      ClassName
      CodeName. 
  
  /** temptable defs for the indicies used by the ObjectConfig class */
  def temp-table TTIndex no-undo xml-node-name "Index"
    field ClassName     as char
    field IndexName      as char
    field ParametersIn   as char
    field ParametersCall as char
    field FindString     as char 

    index Main is primary 
      ClassName
      IndexName. 
     
  /** temptable defs for the properties used by the ObjectConfig class */
  def temp-table TTProperty no-undo xml-node-name "Property"
    field ClassName      as char
    field FieldName      as char 
    field PropertyName   as char 
    field Sequence       as char 
    field DataType       as char 
    field AccessMode     as char 
    field GetMode        as char 
    field SetMode        as char 
    field GetCode        as char 
    field SetCode        as char 
    field Brief          as char 
    field Comment        as char 
    field Fluent         as char
    field UIName         as char
    field UIType         as char
    field FormLabel      as char
    field PropertyGUID   as char
    field fkClassName    as char
    field fkModelName    as char
    field fkPropertyName as char
    
    field extent          as  int  
    field DisplayInLookup as int
    field DisplayInForm   as int
    field DisplayInGrid   as int
    field WidthScale      as int
    
    field GenerateGUID    as logical 
    
    field BuildProperty    as logical 
    field ManualProperty   as logical 
    field OVERRIDE         as logical 
    field checked          as logical
    field isForeignKey     as logical
    field LoadfkProperty   as logical
    field fkDelete         as logical
    field fkUpdate         as logical
    field fkGetAll         as logical
    
    index Main is primary 
      ClassName
      DataType
      PropertyName

    index NAME 
      ClassName
      PropertyName
      
    index FieldName 
      ClassName
      FieldName.

  /** dataset so that we can read and write xml */          
  def dataset dsConfig xml-node-name "Data" for TTObjectConfig, TTProperty, TTCustomCode, TTIndex, TTTemplate
    data-relation for TTObjectConfig, TTProperty   relation-fields (TTObjectConfig.ClassName, TTProperty.ClassName)   nested
    data-relation for TTObjectConfig, TTIndex      relation-fields (TTObjectConfig.ClassName, TTIndex.ClassName)      nested
    data-relation for TTObjectConfig, TTCustomCode relation-fields (TTObjectConfig.ClassName, TTCustomCode.ClassName) nested
    data-relation for TTObjectConfig, TTTemplate   relation-fields (TTObjectConfig.ClassName, TTTemplate.ClassName) nested.

  /** takes an object name and returns an ObjectConfig model 
   * @param p_ClassName : name of object to load
   * @return dotr.Maia.Model.ObjectConfig : ObjectConfig model 
   */

  method public dotr.Maia.Model.ObjectConfig Get(p_ClassName as char):
    def var ObjectConfig1 as dotr.Maia.Model.ObjectConfig no-undo.
    
    def var fk1 as dotr.Maia.Model.fkModel no-undo.

    def var fkObject   as dotr.Maia.Model.ObjectConfig no-undo. 
    def var fkProperty as dotr.Maia.Model.Property     no-undo. 
    
    if p_ClassName eq ? then undo, throw new AppError("Cannot get using an unknown value",0).
    
    def var lv_i     as int no-undo.
    
    def var lv_Data as longchar no-undo.
    
    /** check to see if it the object name, or the actual filename itself that has been passed in */
    assign file-info:file-name = if p_ClassName matches ("*.xml") 
                                    then p_ClassName
                                    else dotr.Maia.CurrentProject:GetConfigFile(p_ClassName).
    
    ObjectConfig1 = new dotr.Maia.Model.ObjectConfig().

    if file-info:full-pathname eq ? then return ObjectConfig1.
    
    fk1 = new dotr.Maia.Model.fkModel().
    
    copy-lob from file file-info:full-pathname to lv_Data.
                              
    /** read the xml file into the temp-tables */
    dataset dsConfig:READ-XML("longchar",lv_Data,"EMPTY",?,?).
    
    find first TTObjectConfig no-lock no-error.
    
    if not avail TTObjectConfig then return ObjectConfig1.

    /** set properties of ObjectConfig */
    assign ObjectConfig1:BuildObject      = TTObjectConfig.BuildObject
           ObjectConfig1:isAbstract       = TTObjectConfig.isAbstract
           ObjectConfig1:ClassName        = TTObjectConfig.ClassName
           ObjectConfig1:ModelName        = if TTObjectConfig.ModelName    eq "" then TTObjectConfig.ClassName                 
                                                                                 else TTObjectConfig.ModelName 
           ObjectConfig1:ModelPackage     = if TTObjectConfig.ModelPackage eq "" then substitute("&1.Model",dotr.Maia.CurrentProject:Name) 
                                                                                 else TTObjectConfig.ModelPackage 
           ObjectConfig1:TableName        = if TTObjectConfig.TableName    eq "" then TTObjectConfig.ClassName
                                                                                 else TTObjectConfig.TableName                                          
           ObjectConfig1:CRC              = TTObjectConfig.CRC
           ObjectConfig1:ConfigFile       = file-info:full-pathname
           ObjectConfig1:KeyFieldDatatype = TTObjectConfig.KeyFieldDatatype 
           ObjectConfig1:KeyFieldName     = TTObjectConfig.KeyFieldName 
           ObjectConfig1:LogicalDb        = TTObjectConfig.LogicalDB 
           ObjectConfig1:KeyProperty      = TTObjectConfig.KeyProperty.
           
    fk1:Removefk(ObjectConfig1:ClassName).
               
    /** loop through all property temp-table records and create a property object */
    for each TTProperty no-lock by TTProperty.ClassName by TTProperty.DataType by TTProperty.PropertyName lv_i = 1 to lv_i + 1 :
      ObjectConfig1:Property[lv_i] = new dotr.Maia.Model.Property().
      
      assign ObjectConfig1:Property[lv_i]:BuildProperty   = TTProperty.BuildProperty
             ObjectConfig1:Property[lv_i]:Brief           = TTProperty.Brief
             ObjectConfig1:Property[lv_i]:Comment         = TTProperty.Comment
             ObjectConfig1:Property[lv_i]:DataType        = TTProperty.DataType
             ObjectConfig1:Property[lv_i]:extent          = TTProperty.extent
             ObjectConfig1:Property[lv_i]:FieldName       = TTProperty.FieldName
             ObjectConfig1:Property[lv_i]:GenerateGUID    = TTProperty.GenerateGUID
             ObjectConfig1:Property[lv_i]:ManualProperty  = TTProperty.ManualProperty
             ObjectConfig1:Property[lv_i]:ClassName       = ObjectConfig1:ModelName
             ObjectConfig1:Property[lv_i]:PropertyName    = TTProperty.PropertyName
             ObjectConfig1:Property[lv_i]:Sequence        = TTProperty.Sequence
             ObjectConfig1:Property[lv_i]:Override        = TTProperty.Override
             ObjectConfig1:Property[lv_i]:AccessMode      = TTProperty.AccessMode
             ObjectConfig1:Property[lv_i]:SetMode         = TTProperty.SetMode
             ObjectConfig1:Property[lv_i]:GetMode         = TTProperty.GetMode
             ObjectConfig1:Property[lv_i]:SetCode         = TTProperty.SetCode
             ObjectConfig1:Property[lv_i]:Fluent          = TTProperty.Fluent
             ObjectConfig1:Property[lv_i]:UIName          = TTProperty.UIName
             ObjectConfig1:Property[lv_i]:UIType          = TTProperty.UIType
             ObjectConfig1:Property[lv_i]:PropertyGUID    = if TTProperty.PropertyGUID eq "" then guid(generate-uuid) else TTProperty.PropertyGUID
             ObjectConfig1:Property[lv_i]:FormLabel       = if TTProperty.FormLabel eq "" then TTProperty.PropertyName
                                                                                          else TTProperty.FormLabel
             ObjectConfig1:Property[lv_i]:DisplayInLookup    = TTProperty.DisplayInLookup
             ObjectConfig1:Property[lv_i]:DisplayInForm      = TTProperty.DisplayInForm
             ObjectConfig1:Property[lv_i]:DisplayInGrid      = TTProperty.DisplayInGrid
             ObjectConfig1:Property[lv_i]:WidthScale         = if TTProperty.WidthScale eq 0 then 1 else TTProperty.WidthScale
             ObjectConfig1:Property[lv_i]:isForeignKey       = TTProperty.isForeignKey
             ObjectConfig1:Property[lv_i]:fkDelete           = TTProperty.fkDelete
             ObjectConfig1:Property[lv_i]:fkUpdate           = TTProperty.fkUpdate
             ObjectConfig1:Property[lv_i]:fkGetAll           = TTProperty.fkGetAll
             ObjectConfig1:Property[lv_i]:LoadfkProperty     = TTProperty.LoadfkProperty
             ObjectConfig1:Property[lv_i]:fkClassName        = TTProperty.fkClassName
             ObjectConfig1:Property[lv_i]:fkModelName        = if TTProperty.fkModelName eq "" then TTProperty.fkClassName else TTProperty.fkModelName
             ObjectConfig1:Property[lv_i]:fkPropertyName     = TTProperty.fkPropertyName
    
             ObjectConfig1:Property[lv_i]:GetCode         = TTProperty.GetCode.

      if ObjectConfig1:Property[lv_i]:AccessMode eq "" or ObjectConfig1:Property[lv_i]:AccessMode eq ? 
         then assign ObjectConfig1:Property[lv_i]:AccessMode = "public".

      if ObjectConfig1:Property[lv_i]:SetMode eq "" or ObjectConfig1:Property[lv_i]:SetMode eq ? 
         then assign ObjectConfig1:Property[lv_i]:SetMode = "public".

      if ObjectConfig1:Property[lv_i]:GetMode eq "" or ObjectConfig1:Property[lv_i]:GetMode eq ? 
         then assign ObjectConfig1:Property[lv_i]:GetMode = "public".

      if ObjectConfig1:Property[lv_i]:isForeignKey then fk1:Addfk(ObjectConfig1:Property[lv_i]:fkClassName,ObjectConfig1:ClassName,"").                  
    
      if ObjectConfig1:Property[lv_i]:isForeignKey and
         ObjectConfig1:Property[lv_i]:fkDelete then fk1:Addfk(ObjectConfig1:Property[lv_i]:fkClassName,ObjectConfig1:ClassName,"delete").                  
    
      if ObjectConfig1:Property[lv_i]:isForeignKey and
         ObjectConfig1:Property[lv_i]:fkUpdate then fk1:Addfk(ObjectConfig1:Property[lv_i]:fkClassName,ObjectConfig1:ClassName,"update").                  

      if ObjectConfig1:Property[lv_i]:isForeignKey and
         ObjectConfig1:Property[lv_i]:fkGetall then fk1:Addfk(ObjectConfig1:Property[lv_i]:fkClassName,ObjectConfig1:ClassName,"getall").
         
      if ObjectConfig1:Property[lv_i]:isForeignKey then
      do:
        fkObject   = dotr.Maia.Model.ObjectConfig:Get(ObjectConfig1:Property[lv_i]:fkClassName).
        fkProperty = new dotr.Maia.Model.Property().
      
        assign fkProperty:BuildProperty   = yes
               fkProperty:AccessMode      = "public"
               fkProperty:DataType        = fkObject:ModelPackage + "." + fkObject:ModelName
               fkProperty:ManualProperty  = yes
               fkProperty:ClassName       = fkObject:ModelName
               fkProperty:PropertyName    = if ObjectConfig1:Property[lv_i]:fkPropertyName eq "" then fkObject:ClassName
                                                                                                 else ObjectConfig1:Property[lv_i]:fkPropertyName 
               fkProperty:PropertyGUID    = guid(generate-uuid).

        assign ObjectConfig1:Property[lv_i]:ForeignKey = fkProperty.
      end.             
    end.
    
    fk1:Save().

    assign ObjectConfig1:NumProperties = lv_i.                 

    /** loop through all custom code temp-table records and create a customcode object */
    for each TTCustomCode no-lock by TTCustomCode.ClassName by TTCustomCode.CodeName lv_i = 1 to lv_i + 1 :
      ObjectConfig1:CustomCode[lv_i] = new dotr.Maia.Model.CustomCode().
      
      assign ObjectConfig1:CustomCode[lv_i]:CodeName = TTCustomCode.CodeName
             ObjectConfig1:CustomCode[lv_i]:Data     = TTCustomCode.Data.
    end.    

    assign ObjectConfig1:NumCustomCode = lv_i.
    
    /** loop through all index temp-table records and create an index object */
    for each TTIndex no-lock by TTIndex.ClassName by TTIndex.IndexName lv_i = 1 to lv_i + 1 :
      ObjectConfig1:Index[lv_i] = new dotr.Maia.Model.Index().
      
      assign ObjectConfig1:Index[lv_i]:IndexName      = TTIndex.IndexName
             ObjectConfig1:Index[lv_i]:parametersIn   = TTIndex.parametersIn
             ObjectConfig1:Index[lv_i]:parametersCall = TTIndex.parametersCall
             ObjectConfig1:Index[lv_i]:FindString     = TTIndex.FindString.
    end.    

    assign ObjectConfig1:NumIndex = lv_i.                 

    /** loop through all property temp-table records and create a property object */
    for each TTTemplate no-lock by TTTemplate.ClassName by TTTemplate.TemplateName lv_i = 1 to lv_i + 1 :
      ObjectConfig1:Template[lv_i] = new dotr.Maia.Model.Template().
      
      assign ObjectConfig1:Template[lv_i]:TemplateName  = TTTemplate.TemplateName
             ObjectConfig1:Template[lv_i]:Build         = TTTemplate.buildTemplate
             ObjectConfig1:Template[lv_i]:BuildIfSame   = TTTemplate.buildIfSame
             ObjectConfig1:Template[lv_i]:BuildIfExists = TTTemplate.buildIfExists
             ObjectConfig1:Template[lv_i]:MD5Hash       = TTTemplate.MD5Hash.
    end.    

    assign ObjectConfig1:NumTemplates = lv_i.                 

    ClearTable().
                     
    return ObjectConfig1.                     
  end.

  /** save ObjectConfig to xml file
   * @param p_ObjectConfig : objectConfig to save
   */

  method public void Save(p_ObjectConfig as dotr.Maia.Model.ObjectConfig):
    WriteRecord(p_ObjectConfig,"CREATE").
  end method.

  /** remove ObjectConfig 
   * @param p_ObjectConfig : objectConfig to remove
   */
  method public void Remove(p_ObjectConfig as dotr.Maia.Model.ObjectConfig):
    WriteRecord(p_ObjectConfig,"Remove").
  end method. 
  
  /** method to create or remove ObjectConfig to xml file
   * @param p_ObjectConfig : objectConfig to save or remove
   * @param p_mode : remove or create
   */

  method private void WriteRecord(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_Mode as char):
    def var lv_i as int no-undo.
    
    def var lv_Data as longchar no-undo.    
    
    def var lv_File as char no-undo.
    
    repeat while true stop-after 5 transaction on error undo, throw
                                               on stop undo, leave:
      case p_Mode:
        when "Create" then
        do:
          
          /** take the ObjectConfig instance and convert into temp-tables */
          ClearTable().
          
          create TTObjectConfig.
          assign TTObjectConfig.ClassName       = p_ObjectConfig:ClassName
                 TTObjectConfig.TableName        = p_ObjectConfig:TableName
                 TTObjectConfig.ModelName        = p_ObjectConfig:ModelName
                 TTObjectConfig.ModelPackage     = p_ObjectConfig:ModelPackage
                 TTObjectConfig.CRC              = p_ObjectConfig:CRC
                 TTObjectConfig.KeyFieldDatatype = p_ObjectConfig:KeyFieldDatatype
                 TTObjectConfig.KeyFieldName     = p_ObjectConfig:KeyFieldName
                 TTObjectConfig.KeyProperty      = p_ObjectConfig:KeyProperty
                 TTObjectConfig.isAbstract       = p_ObjectConfig:isAbstract
                 TTObjectConfig.LogicalDB        = p_ObjectConfig:LogicalDb
                 TTObjectConfig.BuildObject      = p_ObjectConfig:BuildObject.

          do lv_i = 1 to p_ObjectConfig:NumProperties:
            if not valid-object(p_ObjectConfig:Property[lv_i]) or p_ObjectConfig:Property[lv_i]:PropertyName eq "" then next.
            
            create TTProperty.
            assign TTProperty.ClassName       = p_ObjectConfig:ClassName    
                   TTProperty.BuildProperty    = p_ObjectConfig:Property[lv_i]:BuildProperty    
                   TTProperty.Comment          = p_ObjectConfig:Property[lv_i]:Comment    
                   TTProperty.Brief            = p_ObjectConfig:Property[lv_i]:Brief    
                   TTProperty.DataType         = p_ObjectConfig:Property[lv_i]:DataType     
                   TTProperty.extent           = p_ObjectConfig:Property[lv_i]:extent       
                   TTProperty.FieldName        = p_ObjectConfig:Property[lv_i]:FieldName    
                   TTProperty.GenerateGUID     = p_ObjectConfig:Property[lv_i]:GenerateGUID 
                   TTProperty.ManualProperty   = p_ObjectConfig:Property[lv_i]:ManualProperty  
                   TTProperty.Sequence         = p_ObjectConfig:Property[lv_i]:Sequence 
                   TTProperty.AccessMode       = p_ObjectConfig:Property[lv_i]:AccessMode
                   TTProperty.Fluent           = p_ObjectConfig:Property[lv_i]:Fluent
                   TTProperty.GetMode          = p_ObjectConfig:Property[lv_i]:GetMode
                   TTProperty.SetMode          = p_ObjectConfig:Property[lv_i]:SetMode
                   TTProperty.GetCode          = p_ObjectConfig:Property[lv_i]:GetCode
                   TTProperty.SetCode          = p_ObjectConfig:Property[lv_i]:SetCode
                   TTProperty.UIName           = p_ObjectConfig:Property[lv_i]:UIName
                   TTProperty.Formlabel        = p_ObjectConfig:Property[lv_i]:FormLabel
                   TTProperty.PropertyGUID     = p_ObjectConfig:Property[lv_i]:PropertyGUID
                   TTProperty.DisplayInLookup  = p_ObjectConfig:Property[lv_i]:DisplayInLookup
                   TTProperty.DisplayInForm    = p_ObjectConfig:Property[lv_i]:DisplayInForm
                   TTProperty.DisplayInGrid    = p_ObjectConfig:Property[lv_i]:DisplayInGrid
                   TTProperty.UIType           = p_ObjectConfig:Property[lv_i]:UIType
                   TTProperty.WidthScale       = p_ObjectConfig:Property[lv_i]:WidthScale
                   TTProperty.OVERRIDE         = p_ObjectConfig:Property[lv_i]:OVERRIDE
                   TTProperty.isForeignKey     = p_ObjectConfig:Property[lv_i]:isForeignKey       
                   TTProperty.fkDelete         = p_ObjectConfig:Property[lv_i]:fkDelete    
                   TTProperty.fkUpdate         = p_ObjectConfig:Property[lv_i]:fkUpdate    
                   TTProperty.fkGetAll         = p_ObjectConfig:Property[lv_i]:fkGetAll    
                   TTProperty.LoadfkProperty   = p_ObjectConfig:Property[lv_i]:LoadfkProperty 
                   TTProperty.fkClassName      = p_ObjectConfig:Property[lv_i]:fkClassName     
                   TTProperty.fkModelName      = p_ObjectConfig:Property[lv_i]:fkModelName     
                   TTProperty.fkPropertyName   = p_ObjectConfig:Property[lv_i]:fkPropertyName     
                   
                   TTProperty.PropertyName    = replace(p_ObjectConfig:Property[lv_i]:PropertyName,"-","").  
                   
                   
            if TTProperty.Fluent eq "" or TTProperty.Fluent eq ? then assign TTProperty.Fluent = TTProperty.PropertyName.                    
          end.

          do lv_i = 1 to p_ObjectConfig:NumCustomCode:
            if not valid-object(p_ObjectConfig:CustomCode[lv_i]) or p_ObjectConfig:CustomCode[lv_i]:CodeName eq "" then next.
            
            create TTCustomCode.
            assign TTCustomCode.ClassName = p_ObjectConfig:ClassName    
                   TTCustomCode.CodeName   = p_ObjectConfig:CustomCode[lv_i]:CodeName    
                   TTCustomCode.Data       = p_ObjectConfig:CustomCode[lv_i]:Data.
          end.     

          do lv_i = 1 to p_ObjectConfig:NumIndex:
            if not valid-object(p_ObjectConfig:Index[lv_i]) or p_ObjectConfig:Index[lv_i]:IndexName eq "" then next.
            
            create TTIndex.
            assign TTIndex.ClassName     = p_ObjectConfig:ClassName    
                   TTIndex.IndexName      = p_ObjectConfig:Index[lv_i]:IndexName    
                   TTIndex.parametersIn   = p_ObjectConfig:Index[lv_i]:parametersIn   
                   TTIndex.FindString     = p_ObjectConfig:Index[lv_i]:FindString   
                   TTIndex.parametersCall = p_ObjectConfig:Index[lv_i]:parametersCall.
          end.     

          do lv_i = 1 to p_ObjectConfig:NumTemplates:
            if not valid-object(p_ObjectConfig:Template[lv_i]) or p_ObjectConfig:Template[lv_i]:TemplateName eq "" then next.
            
            create TTTemplate.
            assign TTTemplate.ClassName    = p_ObjectConfig:ClassName    
                   TTTemplate.TemplateName  = p_ObjectConfig:Template[lv_i]:TemplateName    
                   TTTemplate.buildIfSame   = p_ObjectConfig:Template[lv_i]:BuildIfSame    
                   TTTemplate.buildIfExists = p_ObjectConfig:Template[lv_i]:BuildIfExists    
                   TTTemplate.buildTemplate = p_ObjectConfig:Template[lv_i]:Build    
                   TTTemplate.MD5Hash       = p_ObjectConfig:Template[lv_i]:MD5Hash.   
          end.     
                
          /** write dataset to file */
          dataset dsConfig:WRITE-XML("longchar",lv_Data, yes  ).
          
          ClearTable().
          
          assign lv_File = dotr.Maia.CurrentProject:GetConfigFile(p_ObjectConfig:ClassName).                             
                                 
          copy-lob from lv_Data to file lv_File.
        end.
        
        when "Remove" then
        do:
          /** not yet implemented */
        end.
        
      end case.
      
      leave.
      
      catch e as Progress.Lang.Error :
        case e:GetMessageNum(1):
          when 132 then undo, throw new Progress.Lang.AppError(replace(replace(e:GetMessage(1),"**",""),"(132)",""),132). 
          
          otherwise     undo, throw new Progress.Lang.AppError(e:GetMessage(1),0).
        end case.
      end catch.
    end.
    
    return.
    
  end method.

  /** gets a property object from the name
   * @param p_ObjectConfig : Object to get property from
   * @param p_Property : property to get
   * @return dotr.Maia.Model.Property : the property
   */

  method public dotr.Maia.Model.Property GetProperty(p_ObjectConfig as dotr.Maia.Model.ObjectConfig ,p_Property as char ):
    def var lv_i     as int no-undo.
    
    do lv_i = 1 to p_ObjectConfig:NumProperties:
      if not valid-object(p_ObjectConfig:Property[lv_i]) then next.
      if p_ObjectConfig:Property[lv_i]:PropertyName eq p_Property then return p_ObjectConfig:Property[lv_i].
    end.
    
    return new dotr.Maia.Model.Property().  
  end method.

  /** gets a Template object from the name
   * @param p_ObjectConfig : Object to get Template from
   * @param p_Property : Template to get
   * @return dotr.Maia.Model.Template : the Template
   */

  method public dotr.Maia.Model.Template GetTemplate(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_Template as char ):
    def var lv_i     as int no-undo.
    
    do lv_i = 1 to p_ObjectConfig:NumTemplates:
      if not valid-object(p_ObjectConfig:Template[lv_i]) then next.
      if p_ObjectConfig:Template[lv_i]:TemplateName eq p_Template then return p_ObjectConfig:Template[lv_i].
    end.
    
    return new dotr.Maia.Model.Template().  
  end method.

  /** gets a property object from the fieldname
   * @param p_ObjectConfig : Object to get property from
   * @param p_fieldname : fieldname to get
   * @return dotr.Maia.Model.Property : the property
   */

  method public dotr.Maia.Model.Property GetFieldName(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_FieldName as char ):
    def var lv_i     as int no-undo.
    
    do lv_i = 1 to p_ObjectConfig:NumProperties:
      if not valid-object(p_ObjectConfig:Property[lv_i]) then next.
      if p_ObjectConfig:Property[lv_i]:FieldName eq p_FieldName then return p_ObjectConfig:Property[lv_i].
    end.
    
    return new dotr.Maia.Model.Property().  
  end method.

  /** adds a property object 
   * @param p_ObjectConfig : Object to add property to
   * @param p_Property : property to add
   */

  method public void AddProperty(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_Property as dotr.Maia.Model.Property ):
    def var lv_i     as int no-undo.
    
    do lv_i = 1 to p_ObjectConfig:NumProperties:
      if not valid-object(p_ObjectConfig:Property[lv_i]) or p_ObjectConfig:Property[lv_i]:PropertyName eq "" then 
      do:
        assign p_ObjectConfig:Property[lv_i]              = p_Property
               p_ObjectConfig:Property[lv_i]:PropertyGUID = guid(generate-uuid).
        return.
      end.
    end.
    
    assign p_ObjectConfig:NumProperties                          = p_ObjectConfig:NumProperties + 1
           p_ObjectConfig:Property[p_ObjectConfig:NumProperties] = p_Property
           p_ObjectConfig:Property[lv_i]:PropertyGUID            = guid(generate-uuid).
    
    return .  
  end method.

  /** gets a CustomCode object 
   * custom code can be inserted in 3 ways. 
   * 1) a customcode directive in the actual generated file
   * 2) a customcode entry in the object config file
   * 3) a customcode snippet in the object config directory
   * 
   * @param p_ObjectConfig : Object to get property from
   * @param p_CustomCode : CustomCode to get
   * @return dotr.Maia.Model.CustomCode : the CustomCode
   */

  method public dotr.Maia.Model.CustomCode GetCustomCode(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_CustomCode as char ):
    def var lv_i     as int no-undo.
    def var lv_Start as int no-undo.
    def var lv_End   as int no-undo.
    
    def var lv_Data as longchar no-undo.
    
    def var lv_CustomStart as char no-undo.
    def var lv_CustomEnd   as char no-undo.
    
    def var CustomCode1 as dotr.Maia.Model.CustomCode no-undo.
    
    /* priority:
       #1 customcode bookmark inside actual program 
       #2 customcode definition in config file
       #3 customcode include in config directory */
       
    assign file-info:file-name = p_ObjectConfig:OutputFile.
        
    /** #1 if the output file exists, load it up, and find all customcode directives */
    if file-info:full-pathname ne ? then  
    do:
      copy-lob from file file-info:full-pathname to lv_Data.
      
      assign lv_CustomStart = replace('@CustomCode(name="Start&1").',"&1",p_CustomCode)
             lv_CustomEnd   = replace('@CustomCode(name="End&1").',"&1",p_CustomCode)
             lv_Start       = index(lv_Data,lv_CustomStart) + length(lv_CustomStart) 
             lv_End         = index(lv_Data,lv_CustomEnd,lv_Start).
      
      if lv_Start ne 0 and lv_End ne 0 then 
      do:
        CustomCode1 = new dotr.Maia.Model.CustomCode().
        assign CustomCode1:CodeName = p_CustomCode
               CustomCode1:Data     = substitute("&1~n&2~n&3",lv_CustomStart,trim(substr(lv_Data,lv_Start,(lv_end - 1) - lv_Start)," ~n_t"),lv_CustomEnd).
        return CustomCode1.
      end.             
    end.
    
    /** #2 check all customcode objects */
    do lv_i = 1 to p_ObjectConfig:NumCustomCode: 
      if not valid-object(p_ObjectConfig:CustomCode[lv_i]) then next.
      if p_ObjectConfig:CustomCode[lv_i]:CodeName eq p_CustomCode then return p_ObjectConfig:CustomCode[lv_i].
    end.
      
    /* #3 check for existance of customcode file in the config directory */
    assign file-info:file-name = substitute("&1/&2/&3.i",dotr.Maia.CurrentProject:ConfigDirectory,p_ObjectConfig:ClassName,p_CustomCode).
    
    if file-info:full-pathname eq ? then return new dotr.Maia.Model.CustomCode().
    
    CustomCode1 = new dotr.Maia.Model.CustomCode().
    assign CustomCode1:CodeName = p_CustomCode.
        
    copy-lob from file file-info:full-pathname to CustomCode1:Data.

    if CustomCode1:Data begins "@BuildInfo(Source=" then CustomCode1:Data = substr(CustomCode1:Data,index(CustomCode1:Data,"~n") + 1).
    
    return Customcode1.
                    
  end method.

  /** adds a CustomCode object 
   * @param p_ObjectConfig : Object to add property to
   * @param p_CustomCode : CustomCode to add
   */

  method public void AddCustomCode(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_CustomCode as dotr.Maia.Model.CustomCode ):
    def var lv_i     as int no-undo.
    
    do lv_i = 1 to p_ObjectConfig:NumCustomCode:
      if not valid-object(p_ObjectConfig:CustomCode[lv_i]) or p_ObjectConfig:CustomCode[lv_i]:CodeName eq "" then 
      do:
        assign p_ObjectConfig:CustomCode[lv_i] = p_CustomCode.
        return.
      end.
    end.
    
    assign p_ObjectConfig:NumCustomCode = p_ObjectConfig:NumCustomCode + 1
           p_ObjectConfig:CustomCode[p_ObjectConfig:NumCustomCode] = p_CustomCode.
    
    return .  
  end method.

  /** adds a Template object 
   * @param p_ObjectConfig : Object to add Template to
   * @param p_Template : Template to add
   */

  method public void AddTemplate(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_Template as char):
    def var Template1 as dotr.Maia.Model.Template no-undo.
    
    def var lv_j as int no-undo.
    
    do lv_j = 1 to num-entries(p_Template):
      Template1 = new dotr.Maia.Model.Template().
      Template1:templateName = entry(lv_j,p_Template).
      AddTemplate(p_ObjectConfig,Template1).
     end.
    
  end method.
  
  /** adds a Template object 
   * @param p_ObjectConfig : Object to add Template to
   * @param p_Template : Template to add
   */

  method public void AddTemplate(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_Template as dotr.Maia.Model.Template ):
    def var lv_i     as int no-undo.
    
    do lv_i = 1 to p_ObjectConfig:NumTemplates:

      if p_ObjectConfig:Template[lv_i]:TemplateName eq p_Template:TemplateName then return.
       
      if not valid-object(p_ObjectConfig:Template[lv_i]) or p_ObjectConfig:Template[lv_i]:TemplateName eq "" then 
      do:
        assign p_ObjectConfig:Template[lv_i] = p_Template.
        return.
      end.
    end.
    
    assign p_ObjectConfig:NumTemplates = p_ObjectConfig:NumTemplates + 1
           p_ObjectConfig:Template[p_ObjectConfig:NumTemplates] = p_Template.
    
    return .  
  end method.

  /** gets a Index object from the fieldname
   * @param p_ObjectConfig : Object to get Index from
   * @param p_Index : Index to get
   * @return dotr.Maia.Model.Index : the Index
   */

  method public dotr.Maia.Model.Index GetIndex(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_Index as char ):
    def var lv_i     as int no-undo.
    
    do lv_i = 1 to p_ObjectConfig:NumIndex:
      if not valid-object(p_ObjectConfig:Index[lv_i]) then next.
      if p_ObjectConfig:Index[lv_i]:IndexName eq p_Index then return p_ObjectConfig:Index[lv_i].
    end.
    
    return new dotr.Maia.Model.Index().  
  end method.

  /** adds a Index object 
   * @param p_ObjectConfig : Object to add Index to
   * @param p_Index : Index to add
   */

  method public void AddIndex(p_ObjectConfig as dotr.Maia.Model.ObjectConfig,p_Index as dotr.Maia.Model.Index ):
    def var lv_i     as int no-undo.
    
    do lv_i = 1 to p_ObjectConfig:NumIndex:
      if not valid-object(p_ObjectConfig:Index[lv_i]) or p_ObjectConfig:Index[lv_i]:IndexName eq "" then 
      do:
        assign p_ObjectConfig:Index[lv_i] = p_Index.
        return.
      end.
    end.
    
    assign p_ObjectConfig:NumIndex = p_ObjectConfig:NumIndex + 1
           p_ObjectConfig:Index[p_ObjectConfig:NumIndex] = p_Index.
    
    return .  
  end method.

  /** clears all temp-table record
   */
  
  method private void ClearTable():
    empty temp-table TTObjectConfig.
    empty temp-table TTProperty.
    empty temp-table TTCustomCode.
    empty temp-table TTIndex.

  end method.
      
end class.

